@@ -2,7 +2,7 @@
 #                                                                               #
 #  Name:       RNetCDF-test.R                                                   #
 #                                                                               #
-#  Version:    1.6.2-2                                                          #
+#  Version:    1.7-3                                                          #
 #                                                                               #
 #  Purpose:    Test functions to the NetCDF interface for R.                    #
 #                                                                               #
@@ -34,6 +34,8 @@
 #  ------   ----       -----------                                              #
 #  pm       29/12/10   First implementation                                     #
 #  mw       18/07/12   Test packed variables                                    #
+#  mw       02/09/14   Test 1D character arrays and character scalars           #
+#  mw       05/09/14   Test reading/writing NC_CHAR as raw bytes                #
 #                                                                               #
 #===============================================================================#
 
@@ -55,20 +57,43 @@ library(RNetCDF)
 
 #--Initialize ------------------------------------------------------------------#
 cat("Starting NetCDF tests...\n")
-nccount <- 0
 
-##  Create a new NetCDF dataset and define two dimensions
+testfun <- function(x,y,tally=NULL) {
+  if (is.null(tally)) {
+    tally <- c(pass=0,fail=0)
+  }
+  # Compare numeric values with single precision tolerance:
+  if (isTRUE(all.equal(x,y,tolerance=2^(-23)))) {
+    cat("OK\n")
+    return(tally+c(1,0))
+  } else {
+    cat("failed\n")
+    return(tally+c(0,1))
+  }
+}
+
+##  Create a new NetCDF dataset and define dimensions
 nc <- create.nc("foo.nc")
 
-dim.def.nc(nc, "station", 5)
-dim.def.nc(nc, "time", unlim=TRUE)
-dim.def.nc(nc, "max_string_length", 32)
+nstation <- 5
+ntime <- 2
+nstring <- 32
+nempty <- 0
 
-##  Create three variables, one as coordinate variable
+dim.def.nc(nc, "station", nstation)
+dim.def.nc(nc, "time", ntime)
+dim.def.nc(nc, "max_string_length", nstring)
+dim.def.nc(nc, "empty", unlim=TRUE)
+
+##  Define variables
 var.def.nc(nc, "time", "NC_INT", "time")
 var.def.nc(nc, "temperature", "NC_DOUBLE", c(0,1))
 var.def.nc(nc, "packvar", "NC_BYTE", c("station"))
 var.def.nc(nc, "name", "NC_CHAR", c("max_string_length", "station"))
+var.def.nc(nc, "qcflag", "NC_CHAR", c("station"))
+var.def.nc(nc, "int0", "NC_INT", NA)
+var.def.nc(nc, "char0", "NC_CHAR", NA)
+var.def.nc(nc, "numempty", "NC_FLOAT", c("station","empty"))
 
 ##  Put some missing_value attribute for temperature
 att.put.nc(nc, "temperature", "missing_value", "NC_DOUBLE", -99999.9)
@@ -79,100 +104,108 @@ att.put.nc(nc, "packvar", "add_offset", "NC_DOUBLE", -5)
 
 ##  Define variable values
 mytime        <- c(1:2)
-mytemperature <- c(1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, NA, NA, 9.9)
+mytemperature <- matrix(c(1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, NA, NA, 9.9),ncol=ntime)
 mypackvar     <- seq_len(5)*10-5
 myname        <- c("alfa", "bravo", "charlie", "delta", "echo")
+myqcflag      <- "ABCDE"
+myint0        <- 12345
+mychar0       <- "?"
 
 ##  Put the data
 var.put.nc(nc, "time", mytime, 1, length(mytime))
-var.put.nc(nc, "temperature", mytemperature, c(1,1), c(5,2))
+var.put.nc(nc, "temperature", mytemperature, c(1,1), c(nstation,ntime))
 var.put.nc(nc, "packvar", mypackvar, pack=TRUE)
-var.put.nc(nc, "name", myname, c(1,1), c(32,5))
+var.put.nc(nc, "name", myname, c(1,1), c(nstring,nstation))
+var.put.nc(nc, "qcflag", charToRaw(myqcflag))
+var.put.nc(nc, "int0", myint0)
+var.put.nc(nc, "char0", mychar0)
 
 sync.nc(nc)
 
-#-- Test 1 (read) --------------------------------------------------------------#
-cat("Test 1 ... ")
-
-x <- c(1, 2)
+## Read tests
+cat("Read numeric vector ... ")
+x <- mytime
+dim(x) <- length(x)
 y <- var.get.nc(nc, 0)
+tally <- testfun(x,y)
 
-if(sum(y %in% x) == 2) {
-    cat("OK\n")
-    nccount <- nccount + 1
-} else
-    cat("failed\n")
-
-#-- Test 2 (read) --------------------------------------------------------------#
-cat("Test 2 ... ")
-
-x <- matrix(data=c(1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, NA, NA, 9.9), ncol=2)
+cat("Read numeric matrix ... ")
+x <- mytemperature
 y <- var.get.nc(nc, "temperature")
+tally <- testfun(x,y,tally)
 
-if(sum(y %in% x) == 10) {
-    cat("OK\n")
-    nccount <- nccount + 1
-} else
-    cat("failed\n")
-
-#-- Test 3 (read) --------------------------------------------------------------#
-cat("Test 3 ... ")
-
-x <- c(6.6, 7.7, NA, NA, 9.9)
+cat("Read numeric matrix slice ... ")
+x <- mytemperature[,2]
+dim(x) <- length(x)
 y <- var.get.nc(nc, "temperature", c(NA,2), c(NA,1))
+tally <- testfun(x,y,tally)
 
-if(sum(y %in% x) == 5) {
-    cat("OK\n")
-    nccount <- nccount + 1
-} else
-    cat("failed\n")
+cat("Read numeric matrix empty slice ... ")
+x <- numeric(0)
+dim(x) <- c(0,1)
+y <- var.get.nc(nc, "temperature", c(NA,2), c(0,1),collapse=FALSE)
+tally <- testfun(x,y,tally)
 
-#-- Test 4 (read) --------------------------------------------------------------#
-cat("Test 4 ... ")
+cat("Read numeric scalar ... ")
+x <- myint0
+dim(x) <- 1
+y <- var.get.nc(nc, "int0")
+tally <- testfun(x,y,tally)
 
-x <- c("alfa", "bravo", "charlie", "delta", "echo" )
+cat("Read numeric empty array ... ")
+x <- numeric(0)
+dim(x) <- c(nstation,nempty)
+y <- var.get.nc(nc, "numempty")
+tally <- testfun(x,y,tally)
+
+cat("Read 2D char array ... ")
+x <- myname
+dim(x) <- length(x)
 y <- var.get.nc(nc, "name")
+tally <- testfun(x,y,tally)
 
-if(sum(y %in% x) == 5) {
-    cat("OK\n")
-    nccount <- nccount + 1
-} else
-    cat("failed\n")
-
-#-- Test 5 (read) --------------------------------------------------------------#
-cat("Test 5 ... ")
-
-x <- c("brav", "char")
+cat("Read 2D char slice ... ")
+x <- substring(myname[2:3],1,4)
+dim(x) <- length(x)
 y <- var.get.nc(nc, "name", c(1,2), c(4,2))
+tally <- testfun(x,y,tally)
 
-if(sum(y %in% x) == 2) {
-    cat("OK\n")
-    nccount <- nccount + 1
-} else
-    cat("failed\n")
+cat("Read 2D char slice as raw bytes ... ")
+x <- substring(myname[2:3],1,4)
+dim(x) <- length(x)
+x <- apply(x,MARGIN=1,FUN=charToRaw)
+y <- var.get.nc(nc, "name", c(1,2), c(4,2), rawchar=TRUE)
+tally <- testfun(x,y,tally)
 
-#-- Test 6 (read) --------------------------------------------------------------#
-cat("Test 6 ... ")
-
-x <- c("bravo", "charlie")
+cat("Read 2D char slice as characters ... ")
+x <- myname[2:3]
+dim(x) <- length(x)
 y <- var.get.nc(nc, "name", c(1,2), c(NA,2))
+tally <- testfun(x,y,tally)
 
-if(sum(y %in% x) == 2) {
-    cat("OK\n")
-    nccount <- nccount + 1
-} else
-    cat("failed\n")
+cat("Read empty 2D char array ... ")
+x <- character(0)
+dim(x) <- 0
+y <- var.get.nc(nc, "name", NA, c(0,0),collapse=FALSE)
+tally <- testfun(x,y,tally)
 
-#-- Test 7 (read unpacked) -----------------------------------------------------#
-cat("Test 7 ... ")
+cat("Read 1D char slice ... ")
+x <- substring(myqcflag,2,3)
+dim(x) <- 1
+y <- var.get.nc(nc, "qcflag", c(2), c(2))
+tally <- testfun(x,y,tally)
 
+cat("Read scalar char ... ")
+x <- mychar0
+dim(x) <- 1
+y <- var.get.nc(nc, "char0")
+tally <- testfun(x,y,tally)
+
+cat("Read and unpack numeric array ... ")
+x <- mypackvar
+dim(x) <- length(x)
 y <- var.get.nc(nc, "packvar", unpack=TRUE)
-
-if(isTRUE(all.equal(mypackvar, as.vector(y)))) {
-    cat("OK\n")
-    nccount <- nccount + 1
-} else
-    cat("failed\n")
+tally <- testfun(x,y,tally)
 
 #-- Close file -----------------------------------------------------------------#
 close.nc(nc)
@@ -182,13 +215,7 @@ close.nc(nc)
 #  UDUNITS calendar functions                                                   #
 #-------------------------------------------------------------------------------#
 
-#--Initialize ------------------------------------------------------------------#
-cat("Starting UDUNITS tests...\n")
-utcount <- 0
-
-#-- Test 1 (utcal.nc() - numeric values) ---------------------------------------#
-cat("Test 1 ... ")
-
+cat("utcal.nc - numeric values ...")
 x <- matrix(data=c(1899, 1900, 1900, 1900, 1900, 1900,
                      12,    1,    1,    1,    1,    1,
 		     31,    1,    1,    1,    1,    1,
@@ -196,65 +223,36 @@ x <- matrix(data=c(1899, 1900, 1900, 1900, 1900, 1900,
 		      0,    0,    0,    0,    0,    0,
 		      0,    0,    0,    0,    0,    0),
 	    ncol=6)
+colnames(x) <- c("year","month","day","hour","minute","second")
 y <- utcal.nc("hours since 1900-01-01 00:00:00 +01:00", c(0:5))
+tally <- testfun(x,y,tally)
 
-if(sum(y == x) == 36) {
-    cat("OK\n")
-    utcount <- utcount + 1
-} else
-    cat("failed\n")
-
-#-- Test 2 (utcal.nc() - string values) ----------------------------------------#
-cat("Test 2 ... ")
-
+cat("utcal.nc - string values ...")
 x <- c("1899-12-31 23:00:00", "1900-01-01 00:00:00", "1900-01-01 01:00:00",
        "1900-01-01 02:00:00", "1900-01-01 03:00:00", "1900-01-01 04:00:00")
 y <- utcal.nc("hours since 1900-01-01 00:00:00 +01:00", c(0:5), type="s")
+tally <- testfun(x,y,tally)
 
-if(sum(y == x) == 6) {
-    cat("OK\n")
-    utcount <- utcount + 1
-} else
-    cat("failed\n")
-
-#-- Test 3 (utinvcal.nc() - numeric values) ------------------------------------#
-cat("Test 3 ... ")
-
+cat("utinvcal.nc - numeric values ...")
 x <- 6.416667
 y <- utinvcal.nc("hours since 1900-01-01 00:00:00 +01:00", c(1900,1,1,5,25,0))
+tally <- testfun(x,y,tally)
 
-if(round(x, 6) == round(y, 6)) {
-    cat("OK\n")
-    utcount <- utcount + 1
-} else
-    cat("failed\n")
-
-#-- Test 4 (utinvcal.nc() - string values) -------------------------------------#
-cat("Test 4 ... ")
-
+cat("utinvcal.nc - string values ...")
 x <- 6.416667
 y <- utinvcal.nc("hours since 1900-01-01 00:00:00 +01:00", "1900-01-01 05:25:00")
-
-if(round(x, 6) == round(y, 6)) {
-    cat("OK\n")
-    utcount <- utcount + 1
-} else
-    cat("failed\n")
-
+tally <- testfun(x,y,tally)
 
 #-------------------------------------------------------------------------------#
 #  Overall summary                                                              #
 #-------------------------------------------------------------------------------#
-cat("Totally ", nccount+utcount, "/ 11 tests passed. ")
-
-if(nccount != 7)
-    stop("Some NetCDF tests failed.", call.=FALSE)
-
-if(utcount != 4)
-    stop("Some UDUNITS tests failed.", call.=FALSE)
-
-cat("Package seems to work properly.\n")
+cat("Summary:", tally["pass"], "pass /", tally["fail"], "fail. ")
 
+if (tally["fail"]==0) {
+  cat("Package seems to work properly.\n")
+} else {
+  cat("Some problems were detected.\n")
+}
 
 #===============================================================================#
 
